from .config import universal_load, flatten_dict, folder_to_file, all_equal
from .resolve import resolve
import sys
import os
import io
import csv

def generate_ascii_table(data):
    # Extracting headers and ensuring consistent row ordering
    headers = ['keys'] + list(data.keys())
    rows = list(next(iter(data.values())).keys())

    # Calculating column widths
    column_widths = [max(len(row) for row in rows)]
    for header in headers[1:]:
        column_width = max(len(header), max(len(str(data[header][row])) for row in rows))
        column_widths.append(column_width)

    # Create the header row
    header_row = " | ".join(headers[i].ljust(column_widths[i]) for i in range(len(headers)))
    separator = "-+-".join("-" * width for width in column_widths)

    # Create data rows
    data_rows = []
    for row in rows:
        row_data = [row] + [str(data[header][row]) if header in data and row in data[header] else '...' for header in headers[1:]]
        data_row = " | ".join(row_data[i].ljust(column_widths[i]) for i in range(len(row_data)))
        data_rows.append(data_row)

    # Combine all parts
    table = [header_row, separator] + data_rows
    return "\n".join(table)

def compare_configs(config_files, output_format=None, path_prefix=''):
    configs = []

    for name in config_files:
        # name is file name
        # first, try to look the name up in the current directory
        if os.path.exists(name):
            name = folder_to_file(name)
            print(f"Selected {name} relative to the current directory.")
        elif os.path.exists(os.path.join(path_prefix, name)):
            name = folder_to_file(os.path.join(path_prefix, name))
            print(f"Selected {name} from the path prefix.")
        else:
            # try to find the name in the storage
            resolved_name = resolve(os.path.join(path_prefix, name))
            if resolved_name is None:
                print(f"Config file {name} not found.")
                continue
            name = folder_to_file(resolved_name)
            print(f"Selected {name} from the storage.")
        config = universal_load(name)
        if config is None:
            print(f"Config file {name} not found or not parsed.")
            continue
        flattened_config = flatten_dict(config, sep='--')
        configs.append(flattened_config)

    keys = set()
    table = {}

    for config in configs:
        keys.update(config.keys())

    for k in keys:
        values = [config.get(k) for config in configs]
        if k == "_from_file":
            continue
        if all_equal(values):
            continue
        for config in configs:
            model_name = os.path.basename(os.path.dirname(config["_from_file"]))
            if model_name not in table:
                table[model_name] = {}
            table[model_name][k] = config.get(k, "...")

    if output_format is None:
        return table
    elif output_format == "json":
        import json
        result = json.dumps(table)
        return result
    elif output_format == "csv":
        output = io.StringIO()
        writer = csv.writer(output)
        for k in keys:
            row = [k]
            strs = [str(config.get(k, "...")) for config in configs]
            if all_equal(strs):
                continue
            row.extend(strs)
            writer.writerow(row)
        return output.getvalue()
    elif output_format == "ascii":
        result = generate_ascii_table(table)
        return result

def main():
    import argparse
    parser = argparse.ArgumentParser(description="Compare configuration files and output the result in ASCII, CSV, or JSON format.")
    parser.add_argument('config_files', nargs='+', help='List of configuration files to compare.')
    parser.add_argument('--csv', action='store_true', help='Output in CSV format.')
    parser.add_argument('--json', action='store_true', help='Output in JSON format.')
    parser.add_argument('--path', type=str, default='', help="Prefix to append to the file names if file wasn't found relative to the current directory")
    args = parser.parse_args()

    if args.csv and args.json:
        parser.error("Please choose either --csv or --json, not both.")

    output_format = "ascii"
    if args.csv:
        output_format = "csv"
    elif args.json:
        output_format = "json"

    result = compare_configs(args.config_files, output_format=output_format, path_prefix=args.path)
    print(result)

if __name__ == "__main__":
    main()



import os
import sys
import json
import yaml
import pickle
import glob
from itertools import groupby

def universal_load(file_path):
    if not os.path.exists(file_path):
        return None
    if file_path.endswith('.json'):
        with open(file_path) as f:
            data = json.load(f)
    elif file_path.endswith('.yaml') or file_path.endswith('.yml'):
        with open(file_path) as f:
            data = yaml.safe_load(f)
    elif file_path.endswith('.pkl'):
        with open(file_path, 'rb') as f:
            data = pickle.load(f)
    else:
        raise ValueError("Unsupported file format")

    data['_from_file'] = file_path
    return data

def universal_store(file_path, data):
    if file_path.endswith('.json'):
        with open(file_path, 'w') as f:
            json.dump(data, f)
    elif file_path.endswith('.yaml') or file_path.endswith('.yml'):
        with open(file_path, 'w') as f:
            yaml.safe_dump(data, f)
    elif file_path.endswith('.pkl'):
        with open(file_path, 'wb') as f:
            pickle.dump(data, f)
    else:
        raise ValueError("Unsupported file format")


def all_equal(iterable):
    g = groupby(iterable)
    return next(g, True) and not next(g, False)

def flatten_dict(d, parent_key='', sep=':'):
    items = []
    for k, v in d.items():
        new_key = f"{parent_key}{sep}{k}" if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        elif isinstance(v, list):
            for i, item in enumerate(v):
                items.extend(flatten_dict({f"{new_key}{sep}{i}": item}, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)

def folder_to_file(fname):
    if os.path.isdir(fname):
        files = glob.glob(fname + "/*.json") + glob.glob(fname + "/*.yml") + glob.glob(fname + "/*.pkl")
        if files:
            fname = files[0]
    return fname


from .resolve import resolve
from .config import universal_load, universal_store

import os
import shutil
import yaml
from collections import OrderedDict

def start_experiment(project_name, experiment_name, config, group="models"):
    # Resolve the path for the experiment
    experiment_path = resolve(f'{group}/{project_name}/{experiment_name}')

    # Ensure the directory exists
    if not os.path.exists(experiment_path):
        os.makedirs(experiment_path)

    # Determine the config type and handle accordingly
    if isinstance(config, dict):
        config_file_path = os.path.join(experiment_path, 'config.yml')
        with open(config_file_path, 'w') as file:
            yaml.safe_dump(config, file)
    elif isinstance(config, str):
        # Assuming config is a file path
        if os.path.exists(config):
            shutil.copy(config, experiment_path)
        else:
            raise FileNotFoundError(f"Config file {config} not found.")
    else:
        raise ValueError("Config must be a dictionary or file path.")
    return experiment_path

def resume_experiment(project_name, experiment_name, config, group="models"):
    # Resolve the path for the experiment
    experiment_path = resolve(f'{group}/{project_name}/{experiment_name}')

    # Ensure the directory and config file exist
    config_file_path = os.path.join(experiment_path, 'config.yml')
    if not os.path.exists(experiment_path) or not os.path.exists(config_file_path):
        raise FileNotFoundError(f"Experiment or config file not found in path: {experiment_path}")

    # Load existing config
    with open(config_file_path, 'r') as file:
        existing_config = yaml.safe_load(file)

    # Compare the existing config with the provided one
    if isinstance(config, dict):
        if config != existing_config:
            raise ValueError("Provided config does not match the existing one.")
    elif isinstance(config, str):
        # Assuming config is a file path
        if os.path.exists(config):
            with open(config, 'r') as file:
                new_config = yaml.safe_load(file)
                if new_config != existing_config:
                    raise ValueError("Provided config file does not match the existing one.")
        else:
            raise FileNotFoundError(f"Config file {config} not found.")
    else:
        raise ValueError("Config must be a dictionary or file path.")
    return experiment_path




import glob
import os
from .config import universal_load, universal_store
import sys
import os
from typing import Callable, Dict, Iterator

# data format comparison for sidecar files:
# json: the worst, because it's hard to append to it
# yaml: ok
# toml: ok
# but text files are bad to store large chunks of data
# pkl: fast, but not human-readable



class Handler:
    def __init__(self, file_path: str, load=True):
        """
        Initializes the Handler with a specific file path.

        Args:
        - file_path (str): The path to the original file.
        """
        self.file_path = file_path
        self.directory, self.base_name = os.path.split(file_path)
        if load:
            self.sidecar_files = self._find()

    def _find(self):
        """
        Looks for sidecar files with any extension and returns a list of the unique identifiers for those sidecar files.

        Returns:
        - list: A list of unique identifiers for the sidecar files found.
        """
        # Collect all files in the directory
        all_files = os.listdir(self.directory)

        # Filter files that match the base name pattern
        matching_files = [file for file in all_files if file.startswith(self.base_name) and file != self.base_name]

        # Extract the unique identifiers from the file names
        unique_identifiers = set()
        for file in matching_files:
            # Extract the part of the file name between the base name and the extension
            identifier = file.replace(f"{self.base_name}---", "")
            unique_identifiers.add(identifier)

        return list(unique_identifiers)

    def get(self, identifier: str):
        """
        Loads the metadata from the sidecar file with the given identifier.

        Args:
        - identifier (str): The unique identifier for the sidecar file.

        Returns:
        - dict: The metadata from the sidecar file with the given identifier.
        """
        sidecar_file_path = os.path.join(self.directory, f"{self.base_name}---{identifier}")

        # Load the metadata from the sidecar file
        metadata = universal_load(sidecar_file_path)

        return metadata

    def get_all(self):
        """
        Loads the metadata from all sidecar files.

        Returns:
        - dict: A dictionary of metadata from all sidecar files.
        """
        all_metadata = {}
        for identifier in self.sidecar_files:
            metadata = self.get(identifier)
            all_metadata[identifier] = metadata

def list_directory(directory: str):
    all_files = set(os.listdir(directory))
    non_sidecar_files = set()
    sidecar_files = {}

    # Organize files into source and sidecar categories
    for file in all_files:
        if '---' in file:
            base_file, _, _ = file.rpartition('---')
            if base_file in all_files:
                sidecar_files.setdefault(base_file, []).append(file)
            else:
                non_sidecar_files.add(file)
        else:
            non_sidecar_files.add(file)
    return non_sidecar_files, sidecar_files

def load_to_pandas(directory: str, identifier: str):
    """
    Loads the data for the whole directory and returns a pandas dataframe
    """
    import pandas as pd

    # Check if the given identifier is a directory
    if not os.path.isdir(directory):
        sys.stderr.write(f"Error: The provided identifier '{identifier}' is not a directory.\n")
        return pd.DataFrame()  # Return an empty DataFrame if the identifier is not a directory

    # Use load_directory to organize files and then process sidecar files
    all_metadata = []
    non_sidecar_files, sidecar_files = load_directory(directory)

    for base_file, sidecar_file_list in sidecar_files.items():
        sidecar_handler = Handler(directory, base_file, load=False)
        for sidecar_file in sidecar_file_list:
            _, _, identifier = os.path.basename(sidecar_file).rpartition('---')
            metadata = sidecar_handler.get(identifier)
            if metadata:
                all_metadata.append(metadata)
            else:
                sys.stderr.write(f"Warning: No {identifier} metadata found for '{sidecar_file}' in directory '{directory}'\n")

    # Convert the list of metadata dictionaries to a pandas DataFrame
    df = pd.DataFrame(all_metadata)

    return df

def update_and_store_sidecar_files(directory: str, identifier: str, function: Callable[[str, Dict], Iterator[Dict]], save_interval: int = 10) -> None:
    """
    Unlike process_sidecar_files, it expects the function to return the updated file and stores it itself
    Kind of wasteful (rewriting the whole file instead of appending), but it will work for most of cases

    Parameters:
    - directory (str): The path to the directory where the non-sidecar and sidecar files are located.
    - identifier (str): A unique identifier used to distinguish between different sidecar files.
    - function (callable): A function that takes (source file path, current state of the sidecar file's data) and returns the updated state.
    - save_interval (int): This parameter controls how often the updated sidecar file data is saved back to disk.
    """
    non_sidecar_files, sidecar_files = list_directory(directory)

    total_files = len(non_sidecar_files)
    for index, file in enumerate(non_sidecar_files):
        print(f"update_and_store_sidecar_files: [{index+1}/{total_files}] {file}")
        file_path = os.path.join(directory, file)
        handler = Handler(file_path, load=False)
        initial_state = handler.get(identifier) or {}
        sidecar_file_path = f"{file_path}---{identifier}"
        updates_generator = function(file_path, initial_state)

        counter = 0
        for updated_state in updates_generator:
            counter += 1
            if counter % save_interval == 0:
                universal_store(sidecar_file_path, updated_state)

        if counter % save_interval != 0:
            universal_store(sidecar_file_path, updated_state)


def process_sidecar_files(directory: str, identifier: str, function: Callable[[str, str], None]) -> None:
    """
    For each non-sidecar file calls the function(file_path, sidecar_file_path)
    It expects the function to check whether the sidecar file exists, fully filled and so on

    Parameters:
    - directory (str): The path to the directory containing the files to process.
    - identifier (str): The identifier used to distinguish sidecar files. Example: "identifier.json"
    - function (callable): A function that takes two arguments (file_path, sidecar_file_path)
    """

    non_sidecar_files, sidecar_files = list_directory(directory)

    for file in non_sidecar_files:
        file_path = os.path.join(directory, file)
        sidecar_file_path = os.path.join(self.directory, f"{file_path}---{identifier}")
        function(file_path, sidecar_file_path)




import os
import sys
import yaml

# Global variable to store the root path
ROOT_PATH = None

def find_config_file(starting_directory):
    """Recursively search upwards from the starting directory to find the configuration file."""
    current_dir = starting_directory
    while True:
        local_config_path = os.path.join(current_dir, 'paths.local.yml')
        default_config_path = os.path.join(current_dir, 'paths.yml')
        if os.path.exists(local_config_path):
            return local_config_path
        elif os.path.exists(default_config_path):
            return default_config_path
        else:
            parent_dir = os.path.dirname(current_dir)
            if parent_dir == current_dir:  # Reached the root of the file system
                break
            current_dir = parent_dir
    return None

def resolve(path):
    """Resolve the path using the root path, initializing it if necessary."""
    global ROOT_PATH
    if ROOT_PATH is None:
        if '__file__' in dir(sys.modules['__main__']):
            entry_point_path = os.path.abspath(sys.modules['__main__'].__file__)
            config_file_path = find_config_file(os.path.dirname(entry_point_path))
            if config_file_path is None:
                config_file_path = find_config_file(os.getcwd())
        else:
            # Handle the case when running from an interpreter or a situation where __main__.__file__ is not set
            config_file_path = find_config_file(os.getcwd())
        if config_file_path:
            with open(config_file_path, 'r') as file:
                try:
                    config = yaml.safe_load(file)
                    ROOT_PATH = config.get('root', None)
                except Exception as e:
                    print(f"Error reading or parsing the configuration file: {e}", file=sys.stderr)
                    ROOT_PATH = None
        if ROOT_PATH is None:
            ROOT_PATH = os.getenv('AIVOL_ROOT_PATH')
        if ROOT_PATH is None:
            ROOT_PATH = "/ai"

    if path.startswith("/"):
        return os.path.join(ROOT_PATH, path.lstrip('/'))
    else:
        return os.path.join(ROOT_PATH, path)